/*   this software is (c) 1997 by Martin Baehr <mbaehr@iaeste.or.at> */

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <getopt.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <WMaker.h>
#include "WMWrapper.h"

#define DEBUG
#ifdef DEBUG
#   define DPRINTF(args) ((void) printf args)
#else
#   define DPRINTF(args) ((void) 0)
#endif


Display *dpy;
Window leader;
WMAppContext *app;
char *wrapclass;
char **wrapcmdline;
int  wrapcmdlinec;

typedef struct cmdline_struct 
{
  char    *display;
  char    *geometry;
  int      start;
  int      hide;
  int      wrapall;
  char    *wrapclass;
  char    *wrapname;
  char   **wrapcmdline;
  int      wrapcmdlinec;
  XContext wrapcontext;
  char    *name;
} cmdline_struct;


static void usage(void)
{
    printf("\nusage: wmwrapper [-help] [-display displayname] [-geometry geom]\n                 [-start] [-wrapall] [-hide] [(<name>|*).](<class>|* \n                 command [command args]\n\n");
}

static pid_t
forkwin(char **argv, int argc, char *geometry)
{
    pid_t pid;

    if ((pid=fork())==0) {
        char **args;
        int i=1;
        if (geometry)
          i+=2; 
        args = malloc(sizeof(char*)*(argc+i));
        if (!args)
          exit(111);
        for (i=0; i<argc; i++) {
            args[i] = argv[i];
        }
        if(geometry)
        {
          DPRINTF(("has geometry:%d:%d\n", argc, i));
          args[i++] = "-geometry";
          args[i++] = geometry;
          DPRINTF(("%s:%s:%s:%d\n", args[i-3], args[i-2], args[i-1], i));
        }
        args[i] = NULL;
        execvp(argv[0], args);
        exit(111);
    }
    return pid;
}

static void 
callback(void *foo, int item, Time time)
{
    DPRINTF(("pushed item %i\n", item));
}

static void
quit(void *foo, int item, Time time)
{
    exit(0);
}



static void
hide(void *foo, int item, Time time)
{
    WMHideApplication(app);
}

Atom delete_win, miniaturize_win;
Atom prots[6];
GNUstepWMAttributes attr;
XWMHints *hints;
WMMenu *menu;
WMMenu *infomenu;
WMMenu *windowmenu;
WMMenu *windowsmenu;
int wincount=0;

static void
wrapwin(Window id, XContext wrapcontext)
{

  XWMHints *hints;
  char *name = NULL;
  char* wrapcontextdata = NULL;
  

//  unwrapwin(id, wrapcontext);
  if(XFindContext(dpy, id, wrapcontext, (XPointer *)&wrapcontextdata) != XCSUCCESS)
  {
    XSaveContext(dpy, id, wrapcontext, "wrap");

    hints = XGetWMHints(dpy, id);
    XFetchName(dpy, id, &name);
    hints->window_group = leader;
    hints->flags |= WindowGroupHint;
    XSetWMHints(dpy, id, hints);
    WMMenuAddItem(windowsmenu, name, (WMMenuAction)callback, NULL, NULL, NULL);
    XFree(name);
  }
}

static void
unwrapwin(Window id, XContext wrapcontext)
{

  XWMHints *hints;

  XDeleteContext(dpy, id, wrapcontext);
   
  hints = XGetWMHints(dpy, id);
  hints->window_group = None;
  hints->flags |= WindowGroupHint;
  XSetWMHints(dpy, id, hints);
}

static void
traversewintree(Window traversewindow, char* wrapname, char* wrapclass, XContext wrapcontext)
{
    Window root, parent;
    Window *children;
    unsigned int nchildren, *nchildproperties;
    int i,j;
    XClassHint childwindow;
    XWMHints *wmhints;


    XQueryTree(dpy, traversewindow, &root, &parent, &children, &nchildren);
    /* first remove all icon windows */
    for (i=0; i<nchildren; i++) {
        if (children[i]==None) 
          continue;

        wmhints = XGetWMHints(dpy, children[i]);
        if (wmhints && (wmhints->flags & IconWindowHint)) {
            for (j = 0; j < nchildren; j++)  {
                if (children[j] == wmhints->icon_window) {
                    children[j] = None;
                    break;
                }
            }
        }
    }

//    DPRINTF(("tree:%s;", wrapclass));

    for (i=0; i<nchildren; i++) 
    {
        if (children[i]==None) 
          continue;
//        if(XGetClassHint(dpy, children[i], &childwindow) && XListProperties(dpy, children[i], nchildproperties) != NULL)
        if(XGetClassHint(dpy, children[i], &childwindow) && XGetWMHints(dpy, children[i]) != NULL)
        {
            DPRINTF(("\ntree:%d:0x%lx; %s, %s, %s.%s\n", i, children[i], childwindow.res_name, childwindow.res_class, wrapname, wrapclass));
            if((wrapclass && wrapname && 
                strcmp(childwindow.res_class, wrapclass) == 0 && 
                strcmp(childwindow.res_name, wrapname) == 0) || 
               (wrapclass && !wrapname && 
                strcmp(childwindow.res_class, wrapclass) == 0) || 
               (!wrapclass && wrapname && 
                strcmp(childwindow.res_name, wrapname) == 0))
            {
                DPRINTF(("trying to wrap from tree: 0x%lx\n", children[i]));
                wrapwin(children[i], wrapcontext);
            }
        }
        traversewintree(children[i], wrapname, wrapclass, wrapcontext);
    } 
    XFree(children);
}

static void
unwrapwintree(Window traversewindow, XContext wrapcontext)
{
    Window root, parent;
    Window *children;
    unsigned int nchildren, *nchildproperties;
    int i,j;
    XClassHint childwindow;
    XWMHints *wmhints;


//    DPRINTF(("in unwrapwintree: 0x%lx\n", traversewindow));
//    DPRINTF(("0x%lx;", traversewindow));
    XQueryTree(dpy, traversewindow, &root, &parent, &children, &nchildren);
//    DPRINTF(("XQueryTree found %d children\n", nchildren));
        
    /* first remove all icon windows */
    for (i=0; i<nchildren; i++)
    {
        if (children[i]==None) 
          continue;

        wmhints = XGetWMHints(dpy, children[i]);
        if (wmhints && (wmhints->flags & IconWindowHint)) 
        {
            for (j = 0; j < nchildren; j++)  
            {
                if (children[j] == wmhints->icon_window) 
                {
                    children[j] = None;
                    break;
                }
            }
        }
    }


    for (i=0; i<nchildren; i++) 
    {
        if (children[i]==None) 
          continue;
        {
            //XPointer wrapcontextdata = NULL;
            char* wrapcontextdata = NULL;
//            DPRINTF(("tree: 0x%lx;", children[i]));
            //if(XFindContext(dpy, children[i], wrapcontext, (XPointer *)&wrapcontextdata) != XCNOENT)
            if(XFindContext(dpy, children[i], wrapcontext, (XPointer *)&wrapcontextdata) == XCSUCCESS)
            {
                DPRINTF(("trying to unwrap from tree: 0x%lx\n", children[i]));
                unwrapwin(children[i], wrapcontext);
            }
        }
//        DPRINTF(("unwrapwintree: 0x%lx\n", children[i]));
        unwrapwintree(children[i], wrapcontext);
    }
    XFree(children);
}

cmdline_struct parse_commandline(int argc, char **argv)
{
    cmdline_struct cmdline;  
    int i,c;
    int digit_optind = 0;

    cmdline.geometry = NULL;
    cmdline.display = NULL;
    cmdline.hide = False;
    cmdline.start = False;
    cmdline.wrapall = False;
    cmdline.wrapname = NULL;
    cmdline.wrapclass = NULL;
    cmdline.wrapcontext = XUniqueContext();
    cmdline.name = "wmwrapper";
    optind = 0;

    if(argc <= 1)
    {
      usage();
      exit(1);
    }

    while (1)
    {
        int this_option_optind = optind ? optind : 1;
        int option_index = 0;
        static struct option long_options[] =
        {
          {"display", 1, 0, 'd'},
          {"geometry", 1, 0, 'g'},
          {"start", 0, 0, 's'},
          {"hide", 0, 0, 'h'},
          {"help", 0, 0, 'H'},
          {"wrapall", 0, 0, 'w'},
          {0, 0, 0, 0}
        };

        c = getopt_long_only (argc, argv, "+",
                   long_options, &option_index);
        if (c == -1)
      break;

        switch (c)
        {
          case 0:
            DPRINTF(("option %s", long_options[option_index].name));
            if (optarg)
              DPRINTF((" with arg %s", optarg));
            DPRINTF(("\n"));
            break;
            ;;
          case 'd':
/* can you tell me why getopt_long_only() is not able to make a difference
   between a required argument and the next option if that argument is 
   missing? */
            if (optarg && strcmp(optarg, "-geometry") && strcmp(optarg, "-start") && strcmp(optarg, "-hide") && strcmp(optarg, "-help") && strcmp(optarg, "-wrapall"))
            {
              DPRINTF(("display: %s\n", optarg));
              cmdline.display = optarg;
            }
            else
            {
              printf("wmwrapper: \"-display\" requires an argument!\n");
              usage();
              exit(1);
            }
            break;
            ;;
          case 'g':
            if (optarg && strcmp(optarg, "-display") && strcmp(optarg, "-start") && strcmp(optarg, "-hide") && strcmp(optarg, "-help") && strcmp(optarg, "-wrapall"))
            {
              DPRINTF(("geometry: %s\n", optarg));
              cmdline.geometry = optarg;
            }
            else
            {
              printf("wmwrapper: \"-geometry\" requires an argument!\n");
              usage();
              exit(1);
            }
            break;
            ;;
          case 's':
            cmdline.start = True;
            break;
            ;;
          case 'h':
            cmdline.hide = True;
            break;
            ;;
          case 'H':
            usage();
            exit(1);
            break;
            ;;
          case 'w':
            cmdline.wrapall = True;
            break;
            ;;
          case '?':
            break;
            ;;
          default:
            DPRINTF(("?? getopt returned character code 0%o ??\n", c));
            ;;
        }
    }

    if (optind < argc)
    {
        cmdline.wrapname = strtok(argv[optind++], ".");
        cmdline.wrapclass = strtok(NULL, ".");
        if(cmdline.wrapclass == NULL)
        {
          cmdline.wrapclass = cmdline.wrapname;
          cmdline.wrapname = NULL;
        }
        if(cmdline.wrapclass && strcmp(cmdline.wrapclass, "*") == 0)
        {
          cmdline.wrapclass = NULL;
        }
        if(cmdline.wrapname && strcmp(cmdline.wrapname, "*") == 0)
        {
          cmdline.wrapname = NULL;   
        }
        if(cmdline.wrapclass == NULL && cmdline.wrapname == NULL)
        {
          printf("what do you expect to get from wrapping all windows?\n");
          usage();
          exit(1);
        }
        if(cmdline.wrapclass == NULL && cmdline.wrapname)
          cmdline.name = cmdline.wrapname;
        if(cmdline.wrapname == NULL && cmdline.wrapclass)
          cmdline.name = cmdline.wrapclass;
        if(cmdline.wrapname && cmdline.wrapclass)
          cmdline.name = argv[optind-1];
        DPRINTF(("name: %s, class: %s\n", cmdline.wrapname, cmdline.wrapclass));

        i=0;
        cmdline.wrapcmdline = malloc(sizeof(char*)*((argc+1)));
        while (optind < argc)
        {
          cmdline.wrapcmdline[i++] = argv[optind++];
        }
        cmdline.wrapcmdline[i] = NULL;
        cmdline.wrapcmdlinec = i;
    }
    else
    {
      printf("name.class argument missing!\n");
      usage();
      exit(1);
    }
    DPRINTF(("return:\n"));
    return cmdline;
}

static void 
mainwin(char *title)
{
    Window win;
    XClassHint classhint;
    XSizeHints            mainSizeHints;

    wincount++;
    mainSizeHints.flags      = PPosition | PSize;    
    mainSizeHints.x          = -300;                    
    mainSizeHints.y          = -300;                    
    mainSizeHints.width      = 1;                
    mainSizeHints.height     = 1;               
    
    win = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 10000, -300, 1, 1, 0, 0, 0);
    XSetWMNormalHints(dpy, win, &mainSizeHints);

    prots[0] = delete_win;
    prots[1] = miniaturize_win;
    XSetWMProtocols(dpy, win, prots, 2);
    XStoreName(dpy, win, title);
    
    /* set class hint */
    classhint.res_name = "wmwrapper";
    classhint.res_class = "WMWrapper";
    XSetClassHint(dpy, win, &classhint);


    hints = XAllocWMHints();
    /* set window group leader */
    hints->window_group = leader;
    hints->flags = WindowGroupHint;
    XSetWMHints(dpy, win, hints);
 
    WMAppAddWindow(app, win);   
    XMapWindow(dpy, win);
}

int main(int argc, char **argv)
{    
    cmdline_struct cmdline;
    XClassHint *classhint;
    char *class;
    int i;
    char *display = "";
    Window wrapselected;

    cmdline.display = getenv("DISPLAY");
    cmdline = parse_commandline(argc, argv);
    dpy = XOpenDisplay(cmdline.display!=NULL?cmdline.display:display);
    if (!dpy) {
	printf("could not open display!");
	exit(1);
    }
    delete_win = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    miniaturize_win = XInternAtom(dpy, "_GNUSTEP_WM_MINIATURIZE_WINDOW", 
				  False);


    leader = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), 0, 0, 10, 10,
			      0, 0, 0);
    XSelectInput(dpy, DefaultRootWindow(dpy), SubstructureNotifyMask); 
    DPRINTF(("leader: 0x%lx, %s\n", leader, cmdline.name));    
    

    /* set class hint */
    classhint = XAllocClassHint();
/*    classhint->res_name = strdup("wmwrapper");*/
    classhint->res_name = strdup(cmdline.name);
/*    class = malloc(strlen("WMWrapper")+strlen(cmdline.wrapclass));*/
    classhint->res_class = "WMWrapper";
/*    classhint->res_class = strcat(strcpy(class, "WMWrapper"), cmdline.wrapclass);*/
    XSetClassHint(dpy, leader, classhint);
    
    /* set window group leader to self */
    hints = XAllocWMHints();
    hints->window_group = leader;
    hints->flags = WindowGroupHint;
    XSetWMHints(dpy, leader, hints);

    /* create app context */
    app = WMAppCreateWithMain(dpy, DefaultScreen(dpy), leader);
    menu = WMMenuCreate(app, cmdline.name);
    infomenu = WMMenuCreate(app, "Info");
    windowmenu = WMMenuCreate(app, "Window");
    windowsmenu = WMMenuCreate(app, "Windows");
    WMMenuAddSubmenu(menu, "Info", infomenu);
    WMMenuAddSubmenu(menu, "Window", windowmenu);
    WMMenuAddSubmenu(menu, "Windows", windowsmenu);
    
    WMMenuAddItem(menu, "Hide", (WMMenuAction)hide, NULL, NULL, NULL);
    WMMenuAddItem(menu, "Quit", (WMMenuAction)quit, NULL, NULL, NULL);
    WMMenuAddItem(infomenu, "Help", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "New", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "Start", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "Wrap All", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "Unwrap All", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "Wrap Selected", (WMMenuAction)callback, NULL, NULL, NULL);
    WMMenuAddItem(windowmenu, "Unwrap Selected", (WMMenuAction)callback, NULL, NULL, NULL);
/*  until i learned how to pass variables via WMMenuAddItem()
    i'll have to use global variables :-( */
    wrapcmdline = cmdline.wrapcmdline;
    wrapcmdlinec = cmdline.wrapcmdlinec;
    WMAppSetMainMenu(app, menu);
    
    WMRealizeMenus(app);

    /* set command to use to startup this */
    XSetCommand(dpy, leader, argv, argc);
    
    /* create first window */
    mainwin(cmdline.name);

    if(cmdline.wrapall)
    {
      XGrabServer(dpy);
      traversewintree(DefaultRootWindow(dpy), cmdline.wrapname, cmdline.wrapclass, cmdline.wrapcontext);
      XUngrabServer(dpy);
    }
    if(cmdline.start)
      forkwin(cmdline.wrapcmdline, cmdline.wrapcmdlinec, cmdline.geometry);
    if(cmdline.hide)
      hide(NULL, 0, 0);
/*    forkwin();   */
 
    XFlush(dpy);
    /*puts("Run xprop on the test window to see the properties defined");*/
    while (1) {
	XEvent ev;
	XNextEvent(dpy, &ev);
        switch(ev.type)
        {
/*	if (ev.type==ClientMessage) { */
        case ClientMessage:
	    if (ev.xclient.data.l[0]==delete_win) {
		XDestroyWindow(dpy,ev.xclient.window);
		wincount--;
	    } else if (ev.xclient.data.l[0]==miniaturize_win) {
                XIconifyWindow(dpy,ev.xclient.window,0);
//	    } else if (ev.xclient.message_type == _XA_WINDOWMAKER_MENU)
	    } 
            else 
            {
                DPRINTF(("ClientMessage: %s: %ld, %ld, %ld", XGetAtomName(dpy, ev.xclient.message_type), ev.xclient.data.l[0], ev.xclient.data.l[1], ev.xclient.data.l[2]));
                switch(ev.xclient.data.l[2])
                {
                  case 108:
                    XGrabServer(dpy);
                    unwrapwintree(DefaultRootWindow(dpy), cmdline.wrapcontext);
                    XUngrabServer(dpy);
                    break;
                    ;;
                  case 109:
                    usage();
                    break;
                    ;;
                  case 110:
                    forkwin(cmdline.wrapcmdline, cmdline.wrapcmdlinec, NULL);
                    break;
                    ;;
                  case 111:
                    forkwin(cmdline.wrapcmdline, cmdline.wrapcmdlinec, cmdline.geometry);
                    break;
                    ;;
                  case 112:
                    XGrabServer(dpy);
                    traversewintree(DefaultRootWindow(dpy), cmdline.wrapname, cmdline.wrapclass, cmdline.wrapcontext);
                    XUngrabServer(dpy);
                    WMRealizeMenus(app); 
                    break;
                    ;;
                  case 113:
                    DPRINTF(("unwrapwintree:\n"));
                    XGrabServer(dpy);
                    unwrapwintree(DefaultRootWindow(dpy), cmdline.wrapcontext);
                    XUngrabServer(dpy);
                    break;
                    ;;
                  case 114:
                    wrapselected = Select_Window(dpy);
                    if(wrapselected != None)
                    {
                      DPRINTF(("trying to wrap selected window: 0x%lx\n", wrapselected));
                      wrapwin(wrapselected, cmdline.wrapcontext);
                      WMRealizeMenus(app); 
                    }
                    break;
                    ;;
                  case 115:
                    wrapselected = Select_Window(dpy);
                    if(wrapselected != None)
                    {
                      DPRINTF(("trying to unwrap selected window: 0x%lx\n", wrapselected));
                      unwrapwin(wrapselected, cmdline.wrapcontext);
                    }
                    break;
                    ;;
                }
            }
            break;
        ;;
/*	}
        if (ev.type==MapNotify) 
        { */
        case CreateNotify:
           DPRINTF(("CreateNotify received "));
           if (ev.xcreatewindow.parent == DefaultRootWindow(dpy) && !ev.xcreatewindow.override_redirect)
           {    
                XClassHint createdwindow;
                Atom property;
                int propertylength;
                XWindowAttributes wattribs;

                DPRINTF(("Window on root created\n"));
                XGetClassHint(dpy, ev.xcreatewindow.window, &createdwindow);
                DPRINTF(("%s, %s, %s.%s", createdwindow.res_name, createdwindow.res_class, cmdline.wrapname, cmdline.wrapclass));
                if((cmdline.wrapclass && cmdline.wrapname && 
                    strcmp(createdwindow.res_class, cmdline.wrapclass) == 0 && 
                    strcmp(createdwindow.res_name, cmdline.wrapname) == 0) || 
                   (cmdline.wrapclass && !cmdline.wrapname && 
                    strcmp(createdwindow.res_class, cmdline.wrapclass) == 0) || 
                   (!cmdline.wrapclass && cmdline.wrapname && 
                    strcmp(createdwindow.res_name, cmdline.wrapname) == 0))
                {
                    DPRINTF(("\nsetting PropertyChangeMask"));
                    XSelectInput(dpy, ev.xcreatewindow.window, PropertyChangeMask);
                }
           }
           DPRINTF(("\n"));
           break;
        ;;
        case PropertyNotify:
           DPRINTF(("PropertyNotify: 0x%lx %s: ", ev.xproperty.window, XGetAtomName(dpy, ev.xproperty.atom)));
           {
                XClassHint createdwindow;
                int propertylength;
                XWindowAttributes wattribs;

                XGetClassHint(dpy, ev.xproperty.window, &createdwindow);
                
                DPRINTF(("%s, %s, %s.%s", createdwindow.res_name, createdwindow.res_class, cmdline.wrapname, cmdline.wrapclass));
                if(((cmdline.wrapclass && cmdline.wrapname && 
                     strcmp(createdwindow.res_class, cmdline.wrapclass) == 0 && 
                     strcmp(createdwindow.res_name, cmdline.wrapname) == 0
                    ) || 
                    (cmdline.wrapclass && !cmdline.wrapname && 
                     strcmp(createdwindow.res_class, cmdline.wrapclass) == 0 
                    ) || 
                    (!cmdline.wrapclass && cmdline.wrapname && 
                     strcmp(createdwindow.res_name, cmdline.wrapname) == 0
                   )) && 
                    ev.xproperty.atom != XInternAtom(dpy, "WM_CLIENT_LEADER", False) && 
                    ev.xproperty.atom != XInternAtom(dpy, "WM_HINTS", False))
                {   
                    DPRINTF(("\ntrying to wrap 0x%lx", ev.xproperty.window));
                    wrapwin(ev.xproperty.window, cmdline.wrapcontext);
                    WMRealizeMenus(app); 
                }
           }
           DPRINTF(("\n"));
            
           break;
        ;;
        }
	WMProcessEvent(app, &ev);
    }
    exit(0);
}

